@@ -3,7 +3,7 @@
 # SPDX-FileCopyrightText: 2018-2021, Marsiske Stefan
 # SPDX-License-Identifier: GPL-3.0-or-later
 
-import sys, os, socket, ssl, struct, platform, getpass
+import sys, os, socket, ssl, struct, platform, getpass, time
 from SecureString import clearmem
 import pysodium
 from qrcodegen import QrCode
@@ -78,8 +78,7 @@ def get_masterkey():
         mk = fd.read()
     return mk
   except FileNotFoundError:
-    print("Error: Could not find masterkey!\nIf sphinx was working previously it is now broken.\nIf this is a fresh install all is good, you just need to run `%s init`." % sys.argv[0])
-    sys.exit(1)
+    raise ValueError("ERROR: Could not find masterkey!\nIf sphinx was working previously it is now broken.\nIf this is a fresh install all is good, you just need to run `%s init`." % sys.argv[0])
 
 def connect():
   ctx = ssl.create_default_context()
@@ -117,7 +116,7 @@ def get_sealkey():
   return sk
 
 def encrypt_blob(blob):
-  # todo implement padding
+  # todo implement padding to hide length information
   sk = get_sealkey()
   nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
   ct = pysodium.crypto_aead_xchacha20poly1305_ietf_encrypt(blob,VERSION,nonce,sk)
@@ -125,7 +124,7 @@ def encrypt_blob(blob):
   return VERSION+nonce+ct
 
 def decrypt_blob(blob):
-  # todo implement padding
+  # todo implement padding to hide length information
   sk = get_sealkey()
   version = blob[:1]
   if version > VERSION:
@@ -147,6 +146,7 @@ def getid(host, user):
   mk = get_masterkey()
   salt = pysodium.crypto_generichash(SALT_CTX, mk)
   clearmem(mk)
+  # todo change this to len(user)|user|len(host)|host
   return pysodium.crypto_generichash(b'|'.join((user.encode(),host.encode())), salt, 32)
 
 def unpack_rule(ct):
@@ -167,17 +167,17 @@ def unpack_rule(ct):
 def pack_rule(char_classes, syms, size, check_digit, xor_mask=None):
   # pack rules into and encrypt them
   if set(char_classes) - {'u','l','d'}:
-    raise ValueError("error: rules can only contain any of 'uld'.")
+    raise ValueError("ERROR: rules can only contain any of 'uld'.")
   if set(syms) - set(bin2pass.symbols) != set():
-    raise ValueError("error: symbols can only contain any of '%s'." % bin2pass.symbols)
+    raise ValueError("ERROR: symbols can only contain any of '%s'." % bin2pass.symbols)
   if xor_mask is None and (char_classes == '' and len(syms)<2):
-    raise ValueError("error: no char classes and not enough symbols specified.")
+    raise ValueError("ERROR: no char classes and not enough symbols specified.")
   if xor_mask is None:
       xor_mask = b'\x00' * 32
   elif len(xor_mask)!=32:
-    raise ValueError("error: xor_mask must be 32bytes, is instead: %d." % len(xor_mask))
+    raise ValueError("ERROR: xor_mask must be 32bytes, is instead: %d." % len(xor_mask))
   if size<0 or size>127:
-    raise ValueError("error: invalid max password size: %d." % size)
+    raise ValueError("ERROR: invalid max password size: %d." % size)
 
   packed = size
   packed = packed + (sum(1<<i for i, c in enumerate(('u','l','d')) if c in char_classes) << 7)
@@ -196,11 +196,10 @@ def commit_undo(s, op, pwd, user, host):
   s = ratelimit(s, msg)
   if not auth(s,id,pwd,r):
     s.close()
-    raise ValueError("auth failed")
+    raise ValueError("Failed to authenticate to server while %s" % "committing" if op == COMMIT else "undoing")
   if s.recv(2)!=b'ok':
-    print("ohoh, something is corrupt, and this is a bad, very bad error message in so many ways", file=sys.stderr)
     s.close()
-    raise ValueError("Operation failed")
+    raise ValueError("Server failed to %s" % "Commit" if op == COMMIT else "UNDO")
   s.close()
   return True
 
@@ -229,7 +228,7 @@ def update_rec(s, host, item): # this is only for user blobs. a UI feature offer
       bsize = len(blob)
       if bsize >= 2**16:
         s.close()
-        raise ValueError("error: blob is bigger than 64KB. %d" % bsize)
+        raise ValueError("ERROR: list of usernames is bigger than 64KB. %d" % bsize)
       blob = struct.pack("!H", bsize) + blob
       # writes need to be signed, and sinces its a new blob, we need to attach the pubkey
       blob = b''.join([pk, blob])
@@ -239,9 +238,9 @@ def update_rec(s, host, item): # this is only for user blobs. a UI feature offer
       blob = read_pkt(s, bsize)
       if blob == b'fail':
         s.close()
-        raise ValueError("reading user blob failed")
+        raise ValueError("reading list of user names failed")
       version, blob = decrypt_blob(blob)
-      items = set(blob.decode().split('\x00'))
+      items = {x for x in blob.decode().split('\x00') if x}
       # this should not happen, but maybe it's a sign of corruption?
       if item in items:
         print(f'warning: "{item}" is already in the user record', file=sys.stderr)
@@ -252,7 +251,7 @@ def update_rec(s, host, item): # this is only for user blobs. a UI feature offer
       bsize = len(blob)
       if bsize >= 2**16:
         s.close()
-        raise ValueError("error: blob is bigger than 64KB. %d" % bsize)
+        raise ValueError("ERROR: list of user names is bigger than 64KB. %d" % bsize)
       blob = struct.pack("!H", bsize) + blob
       blob = sign_blob(blob, id, b'')
     s.send(blob)
@@ -282,11 +281,16 @@ def ratelimit(s,req):
   s.send(pkt0)
   challenge = s.recv(1+1+8+32) # n,k,ts,sig
   if len(challenge)!= 1+1+8+32:
-    print("challengelen incorrect: %s %s" %(len(challenge), repr(challenge)), file=sys.stderr)
-    raise ValueError("failed to get ratelimit challenge")
+    if verbose: print("challengelen incorrect: %s %s" %(len(challenge), repr(challenge)), file=sys.stderr)
+    raise ValueError("ERROR: failed to get ratelimit challenge")
   s.close()
   n = challenge[0]
   k = challenge[1]
+
+  try:
+    os.write(3,f"{n} {k}\n".encode('utf8'))
+  except OSError: pass
+
   if k==4:
     if n < 90:
       if verbose: print("got an easy puzzle: %d" % n, file=sys.stderr)
@@ -295,7 +299,14 @@ def ratelimit(s,req):
     else:
       if verbose: print("got a moderate puzzle: %d" % n, file=sys.stderr)
   seed = challenge + req
+
+  delta = time.time()
   solution = solve(n, k, seed)
+  delta = time.time() - delta
+  try:
+    os.write(3,f"{delta}".encode('utf8'))
+  except OSError: pass
+
   s = connect()
   pkt1 = b''.join([CHALLENGE_VERIFY, challenge])
   s.send(pkt1)
@@ -323,6 +334,9 @@ def init_key():
     with open(kfile,'wb') as fd:
       if not win: os.fchmod(fd.fileno(),0o600)
       fd.write(mk)
+  except:
+    print("ERROR: failed to initialize master key", file=sys.stderr)
+    return 1
   finally:
     clearmem(mk)
   return 0
@@ -338,7 +352,10 @@ def create(s, pwd, user, host, char_classes='uld', symbols=bin2pass.symbols, siz
   beta = s.recv(32)
   if beta == b'\x00\x04fail':
     s.close()
-    raise ValueError("error: sphinx protocol failure.")
+    raise ValueError("ERROR: Creating new password, the record probably already exists or the first message to server was corrupted during transport.")
+    # or (less probable) the initial message was longer/shorter than the 65 bytes we sent
+    # or (even? less probable) the value alpha received by the server is not a valid point
+    # both of these less probable causes point at corruption during transport
   rwd = sphinxlib.finish(pwd, r, beta, id)
 
   # second phase, derive new auth signing pubkey
@@ -385,7 +402,7 @@ def get(s, pwd, user, host):
   resp = s.recv(32+RULE_SIZE) # beta + sealed rules
   if resp == b'\x00\x04fail' or len(resp)!=32+RULE_SIZE:
       s.close()
-      raise ValueError("error: sphinx protocol failure.")
+      raise ValueError("ERROR: Either the record does not exist, or the request to server was corrupted during transport.")
   beta = resp[:32]
   rules = resp[32:]
   rwd = sphinxlib.finish(pwd, r, beta, id)
@@ -394,11 +411,11 @@ def get(s, pwd, user, host):
     classes, symbols, size, checkdigit, xormask = unpack_rule(rules)
   except ValueError:
     s.close()
-    return
+    raise ValueError("ERROR: failed to unpack password rules from server")
   s.close()
 
   if validate_password and (checkdigit != (pysodium.crypto_generichash(CHECK_CTX, rwd, 1)[0] & ((1<<5)-1))):
-    raise ValueError("bad checkdigit")
+    raise ValueError("ERROR: bad checkdigit")
 
   rwd = xor(pysodium.crypto_generichash(PASS_CTX, rwd),xormask)
   ret = bin2pass.derive(rwd,classes,size,symbols)
@@ -434,16 +451,15 @@ def change(s, oldpwd, newpwd, user, host, classes='uld', symbols=bin2pass.symbol
   s = ratelimit(s, msg)
   # auth: do sphinx with current seed, use it to sign the nonce
   if not auth(s,id,oldpwd,r):
-    print('failed authentication', file=sys.stderr)
     s.close()
-    return
+    raise ValueError("ERROR: Failed to authenticate using old password to server while changing password on server or record doesn't exist")
 
   r, alpha = sphinxlib.challenge(newpwd)
   s.send(alpha)
   beta = s.recv(32) # beta
   if beta == b'\x00\x04fail' or len(beta)!=32:
     s.close()
-    raise ValueError("error: sphinx protocol failure.")
+    raise ValueError("ERROR: changing password failed due to corruption during transport.")
   rwd = sphinxlib.finish(newpwd, r, beta, id)
 
   if validate_password:
@@ -452,10 +468,9 @@ def change(s, oldpwd, newpwd, user, host, classes='uld', symbols=bin2pass.symbol
       checkdigit = 0
 
   if target:
-    xormask = xor(pysodium.crypto_generichash(PASS_CTX, rwd),bin2pass.pass2bin(target))
+    trwd, classes, symbols = bin2pass.pass2bin(target, None)
+    xormask = xor(pysodium.crypto_generichash(PASS_CTX, rwd),trwd)
     size = len(target)
-    classes = 'uld'
-    symbols = bin2pass.symbols
   else:
     xormask = pysodium.randombytes(32)
 
@@ -467,9 +482,8 @@ def change(s, oldpwd, newpwd, user, host, classes='uld', symbols=bin2pass.symbol
   s.send(sign_blob(b''.join([pk,rule]), id, rwd))
 
   if s.recv(2)!=b'ok':
-    print("ohoh, something is corrupt, and this is a bad, very bad error message in so many ways", file=sys.stderr)
     s.close()
-    return
+    raise ValueError("ERROR: failed to update password rules on the server during changing of password.")
 
   s.close()
   rwd = xor(pysodium.crypto_generichash(PASS_CTX, rwd),xormask)
@@ -493,7 +507,7 @@ def delete(s, pwd, user, host):
   rwd = auth(s,id,pwd,r)
   if not rwd:
     s.close()
-    return
+    raise ValueError("ERROR: Failed to authenticate to server while deleting password on server or record doesn't exist")
 
   # delete user from user list for this host
   # a malicous server could correlate all accounts on this services to this users here
@@ -506,20 +520,19 @@ def delete(s, pwd, user, host):
   bsize = struct.unpack('!H', bsize)[0]
   if bsize == 0:
     # this should not happen, it means something is corrupt
-    print("error: server has no associated user record for this host", file=sys.stderr)
     s.close()
-    return
+    raise ValueError("ERROR: server has no associated user record for this host", file=sys.stderr)
 
   blob = s.recv(bsize)
-  # todo handle this
   if blob == b'fail':
     s.close()
-    return
+    raise ValueError("ERROR: invalid signature on list of users", file=sys.stderr)
   version, blob = decrypt_blob(blob)
   users = set(blob.decode().split('\x00'))
   if user not in users:
     # this should not happen, but maybe it's a sign of corruption?
-    print(f'warning "{user}" is not in user record', file=sys.stderr)
+    s.close()
+    raise ValueError(f'warning "{user}" is not in user record', file=sys.stderr)
   users.remove(user)
   blob = ('\x00'.join(sorted(users))).encode()
   # notice we do not add rwd to encryption of user blobs
@@ -527,7 +540,7 @@ def delete(s, pwd, user, host):
   bsize = len(blob)
   if bsize >= 2**16:
     s.close()
-    raise ValueError("error: blob is bigger than 64KB.")
+    raise ValueError("ERROR: blob is bigger than 64KB.")
   blob = struct.pack("!H", bsize) + blob
   blob = sign_blob(blob, id, b'')
 
@@ -535,7 +548,7 @@ def delete(s, pwd, user, host):
 
   if b'ok' != s.recv(2):
     s.close()
-    return
+    raise ValueError("ERROR: server failed to save updated list of user names for host: %s." % host)
 
   s.close()
   clearmem(rwd)
@@ -572,7 +585,7 @@ def qrcode(output, key):
   else:
     print(qr.to_svg_str(2))
 
-def usage(params):
+def usage(params, help=False):
   print("usage: %s init" % params[0])
   print("usage: echo -n 'password' | %s <create|change> <user> <site> <[u][l][d][s] [<size>] [<symbols>]> | [<target password>]" % params[0])
   print("usage: echo -n 'password' | %s get <user> <site>" % params[0])
@@ -580,7 +593,8 @@ def usage(params):
   print("usage: echo -n 'password' | %s <commit|undo|delete> <user> <site> # if rwd_keys is true in your config" % params[0])
   print("usage: %s list <site>" % params[0])
   print("usage: %s qr [svg] [key]" % params[0])
-  sys.exit(1)
+  if help: sys.exit(0)
+  sys.exit(100)
 
 def arg_rules(params):
   user = params[2]
@@ -596,6 +610,7 @@ def arg_rules(params):
         classes = ''.join(set(param) - set(['s']))
       else:
         classes = param
+        symbols = ''
       continue
     if not size:
       try:
@@ -612,7 +627,7 @@ def arg_rules(params):
   if target is not None and (symbols or classes or size):
     print(f"invalid args for {param[1]}: \"{params[4:]}\"", file=sys.stderr)
     usage(param)
-  return user, site, classes or 'uld', symbols or bin2pass.symbols, size or 0, target
+  return user, site, classes or 'uld', symbols if symbols is not None else bin2pass.symbols, size or 0, target
 
 def test_pwd(pwd):
   q = zxcvbn(pwd.decode('utf8'))
@@ -627,10 +642,12 @@ def test_pwd(pwd):
 #### main ####
 
 def main(params=sys.argv):
-  if len(params) < 2: usage(params)
+  if len(params) < 2: usage(params, True)
   cmd = None
   args = []
-  if params[1] == 'create':
+  if params[1] in ('help', '-h', '--help'):
+    usage(params, True)
+  elif params[1] == 'create':
     try:
       user,site,classes, syms, size, target = arg_rules(params)
     except: usage(params)
@@ -678,48 +695,60 @@ def main(params=sys.argv):
     if params[2:]: usage(params)
     qrcode(output, key)
     return
-
-  if cmd is not None:
-    if cmd != users:
-      pwd = ''
-      if (rwd_keys or cmd in {create,change,get}):
-        pwd = getpwd()
-        if cmd == change:
-          newpwd = getpwd()
-          if not newpwd:
-            newpwd = pwd
-          test_pwd(newpwd)
-          args=(newpwd,) + args
-        if cmd == create:
-          test_pwd(pwd)
-      try:
-        s = connect()
-        ret = cmd(s, pwd, *args)
-      except:
-        ret = False
-        raise # todo remove only for dbg
-      clearmem(pwd)
-    else:
-      try:
-        s = connect()
-        ret = cmd(s,  *args)
-      except:
-        ret = False
-        raise # todo remove only for dbg
-    s.close() # todo is still needed?
-    if not ret:
-      print("fail", file=sys.stderr)
-      sys.exit(1)
-    if cmd not in {delete, undo, commit}:
-      print(ret)
-      sys.stdout.flush()
-      clearmem(ret)
   else:
-    usage()
+    usage(params)
+
+  error = None
+  s = None
+  if cmd != users:
+    pwd = ''
+    if (rwd_keys or cmd in {create,change,get}):
+      pwd = getpwd()
+      if cmd == change:
+        newpwd = getpwd()
+        if not newpwd:
+          newpwd = pwd
+        test_pwd(newpwd)
+        args=(newpwd,) + args
+      if cmd == create:
+        test_pwd(pwd)
+    try:
+      s = connect()
+      ret = cmd(s, pwd, *args)
+    except Exception as exc:
+      error = exc
+      ret = False
+      #raise # only for dbg
+    clearmem(pwd)
+  else:
+    try:
+      s = connect()
+      ret = cmd(s,  *args)
+    except Exception as exc:
+      error = exc
+      ret = False
+      #raise # only for dbg
+  if s and s.fileno() != -1: s.close()
+
+  if not ret:
+    if not error:
+        print("fail", file=sys.stderr)
+        sys.exit(3) # error not handled by exception
+    print(error, file=sys.stderr)
+    if str(error) == "ERROR: bad checkdigit":
+      sys.exit(2) # bad check digit
+    sys.exit(1) # generic errors
+
+  if cmd not in {delete, undo, commit}:
+    print(ret)
+    sys.stdout.flush()
+    clearmem(ret)
+  elif ret != True:
+    print("reached code that should not be reachable: ", ret)
 
 if __name__ == '__main__':
   try:
     main(sys.argv)
   except Exception:
     print("fail", file=sys.stderr)
-    raise # todo remove only for dbg
+    #raise # only for dbg
